/* 

    TiMidity++ -- MIDI to WAVE converter and player
    Copyright (C) 1999 Masanao Izumo <mo@goice.co.jp>
    Copyright (C) 1995 Tuukka Toivonen <tt@cgs.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	Macintosh interface for TiMidity
	by T.Nogami	<t-nogami@happy.email.ne.jp>
	
    mac_oms.c
    Macintosh OMS handler
*/

#ifdef MAC_USE_OMS
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include "timidity.h"
#include "controls.h"
#include "instrum.h"
#include "playmidi.h"
#include "readmidi.h"
#include "recache.h"
#include "output.h"
#include "aq.h"
#include "timer.h"

#include "utime.h"
#include "mac_oms.h"

#include "soundcard.h"


#define DEFAULT_LOW_TIMEAT	0.4
#define DEFAULT_HIGH_TIMEAT	0.6
#define DEFAULT_TIMEBASE	100 /* HZ? */

typedef struct mac_midi_event{
	struct mac_midi_event*	next;
	int	type;
	union{
	OMSMIDIPacket	packet;
		OMSTimerMessage* timerMessage;
	};
	long		mactime;
} Mac_Midi_Packet;
#define	MIDI_PACKET 1

static int curr_timebase = DEFAULT_TIMEBASE;
static int32 sample_correction;
static int32 sample_increment;
static int32 sample_cum;
static int32 curr_event_samples, event_time_offset;
static int32 curr_tick, tick_offs;
static double start_time;
static int tmr_running;

static int is_system_prefix = 0;
static double low_time_at = 0.3;
static double high_time_at = 0.5;

static int		mac_osm_enable=0;
static volatile int	oms_counter=0;
OMSConnectionParams	connects;
const OMSSignature	inID= 'INP ';
OMSTimerClient		*timer_client;
Mac_Midi_Packet *packet_head, *packet_tail;

// ********************************************************************
#pragma mark -

static void server_reset(void);

static void server_setup()
{
    ctl->flags &= ~(CTLF_LIST_RANDOM|CTLF_LIST_SORT);
    opt_realtime_playing = 2; /* Enable loading patch while playing */
    allocate_cache_size = 0; /* Don't use pre-calclated samples */
/*  aq_set_soft_queue(-1.0, 0.0); */
	server_reset();
}

// ********************************************************************
#pragma mark -

#if 1
static void do_timeout()
{
}

static void seq_play_event(MidiEvent *ev)
{
    /*if(tmr_running)
	ev->time = curr_event_samples;
    else*/
    {
	/*if(IS_STREAM_TRACE)
	{
	    event_time_offset += play_mode->rate / MIDI_COMMAND_PER_SEC;
	    ev->time = curr_event_samples;
	}
	else*/
	{
	    double past_time = get_current_calender_time() - start_time;
	    if(play_mode->flag & PF_PCM_STREAM)
		//past_time += high_time_at;
		past_time -= 20.0; //fix me
	    ev->time = (int32)(past_time * play_mode->rate);
	    //ctl->cmsg(CMSG_ERROR, VERB_NORMAL, "t=%d, ", ev->time );

	}
    }
    //ev->time += event_time_offset;
	//ctl->cmsg(CMSG_ERROR, VERB_NORMAL, "call play_event" );
    play_event(ev);
    //ctl->cmsg(CMSG_ERROR, VERB_NORMAL, "call end play_event" );
}

static void tmr_reset(void)
{
    curr_event_samples =
	event_time_offset =
	    sample_cum = 0;
    playmidi_tmr_reset();
    curr_timebase = DEFAULT_TIMEBASE;
    curr_tick = tick_offs = 0;
    start_time = get_current_calender_time();
}

static void compute_sample_increment(void)
{
    double a;
    a = (double)current_play_tempo * (double)play_mode->rate
	* (65536.0/500000.0) / (double)curr_timebase,
    sample_correction = (int32)(a) & 0xFFFF;
    sample_increment = (int32)(a) >> 16;
}

static void stop_playing(void)
{
    if(upper_voices)
    {
	MidiEvent ev;
	ev.type = ME_EOT;
	ev.a = 0;
	ev.b = 0;
	seq_play_event(&ev);
	aq_flush(0);
    }
}

static void do_sysex(uint8 *, int len);

#if 0
static void doit(void)
{
    //while(control_fd != -1)
    while(1);
    {
	int n, maxfd;

	{
	    long usec;


	    //if(data_fd != -1)
	    {
		if(upper_voices == 0 || !IS_STREAM_TRACE)
		    usec = -1;
		else
		{
		    double wait_time;
		    int32 filled;

		    filled = aq_filled();
		    if(tmr_running && filled <= 0)
			usec = -1;
		    else
		    {
			wait_time = (double)filled / play_mode->rate
			    - low_time_at;
			if(wait_time <= 0)
			    usec = 0;
			else
			    usec = (long)(wait_time * 1000000);
		    }
		}
	    }
	    //else
	//	usec = -1;

	    if(usec >= 0)
	    {
		struct timeval timeout;
		timeout.tv_sec = usec / 1000000;
		timeout.tv_usec = usec % 1000000;
		//n = select(maxfd + 1, &fds, NULL, NULL, &timeout);
	    }
	    //else
		//n = select(maxfd + 1, &fds, NULL, NULL, NULL);

	    if(n < 0)
	    {
		//perror("select");
		break;
	    }

	    if(n == 0)
	    {
		if(ctl->verbosity >= VERB_DEBUG)
		{
		    putchar(',');
		    fflush(stdout);
		}
		do_timeout();
		continue;
	    }

	    if(data_fd != -1 && FD_ISSET(data_fd, &fds))
	    {
		if(do_sequencer())
		{
		    close(data_fd);
		    data_fd = -1;
		    send_status(403, "Data connection is closed");
		}
	    }
	}
    }

    if(data_fd != -1)
	stop_playing();
}
#endif

static void server_reset(void)
{
    playmidi_stream_init();
    if(free_instruments_afterwards)
	free_instruments(0);
    //data_buffer_len = 0;
    do_sysex(NULL, 0); /* Initialize SysEx buffer */
    low_time_at = DEFAULT_LOW_TIMEAT;
    high_time_at = DEFAULT_HIGH_TIMEAT;
    reduce_voice_threshold = 0; /* Disable auto reduction voice */
    compute_sample_increment();
    tmr_reset();
    tmr_running = 0;
    start_time = get_current_calender_time();
}

static void make_event(const uint8 data[], MidiEvent ev)
{
    switch(data[0] & 0xf0)
    {
      case MIDI_NOTEON:
	ev.channel = data[0] & 0x0f;
	ev.a       = data[1] & 0x7f;
	ev.b       = data[2] & 0x7f;
	if(ev.b != 0)
	    ev.type = ME_NOTEON;
	else
	    ev.type = ME_NOTEOFF;
	seq_play_event(&ev);
	break;

      case MIDI_NOTEOFF:
	ev.type    = ME_NOTEOFF;
	ev.channel = data[0] & 0x0f;
	ev.a       = data[1] & 0x7f;
	ev.b       = data[2] & 0x7f;
	seq_play_event(&ev);
	break;

      case MIDI_KEY_PRESSURE:
	ev.type    = ME_KEYPRESSURE;
	ev.channel = data[0] & 0x0f;
	ev.a       = data[1] & 0x7f;
	ev.b       = data[2] & 0x7f;
	seq_play_event(&ev);
	break;

      case MIDI_CTL_CHANGE:
	if(convert_midi_control_change(data[0] & 0x0f,
				       data[1],
				       data[2],
				       &ev))
	    seq_play_event(&ev);
	break;

      case MIDI_PGM_CHANGE:
	ev.type    = ME_PROGRAM;
	ev.channel = data[0] & 0x0f;
	ev.a       = data[1] & 0x7f;
	ev.b       = 0;
	seq_play_event(&ev);
	break;

      case MIDI_CHN_PRESSURE:
	ev.type    = ME_CHANNEL_PRESSURE;
	ev.channel = data[0] & 0x0f;
	ev.a       = data[1] & 0x7f;
	ev.b       = 0;
	seq_play_event(&ev);
	break;

      case MIDI_PITCH_BEND:
	ev.type    = ME_PITCHWHEEL;
	ev.channel = data[0] & 0x0f;
	ev.a       = data[1] & 0x7f;
	ev.b       = data[2] & 0x7f;
	seq_play_event(&ev);
	break;

     // case MIDI_SYSTEM_PREFIX:
	//do_sysex(data_buffer + offset + 1, 1);
	//is_system_prefix = 1; /* Start SysEX */
	//break;

      default:
	 //ctl->cmsg(CMSG_ERROR, VERB_NORMAL,
	//	 "Undefined SEQ_MIDIPUTC 0x%02x",
	//	 data[0]);
	/*send_status(402, "Undefined SEQ_MIDIPUTC 0x%02x",
		    data[0]);*/
	return;
    }
}

static void do_sysex(uint8 *data, int n)
{
    static uint8 sysexbuf[BUFSIZ];
    static int buflen;
    static int fillflag;
    int i;

    if(data == NULL)
    {
	buflen = fillflag = 0;
	is_system_prefix = 0;
	return;
    }

    for(i = 0; i < n; i++)
    {
	/* SysEx := /\xf0([^\xf7]*\xf7)/ */
	if(!fillflag)
	{
	    if(data[i] == 0xf0)
	    {
		fillflag = 1;
		continue;
	    }
	}
	if(fillflag)
	{
	    if(buflen < sizeof(sysexbuf))
		sysexbuf[buflen++] = data[i];
	    if(data[i] == 0xf7)
	    {
		MidiEvent ev;
		if(parse_sysex_event(sysexbuf, buflen, &ev))
		    seq_play_event(&ev);
		buflen = 0;
		fillflag = 0;
	    }
	}
    }
}

static void do_sequencer(const Mac_Midi_Packet* p)
{
    int n, offset;
    MidiEvent ev;
	/*ctl->cmsg(CMSG_ERROR, VERB_NORMAL,
		"len=%d, %02X %02X %02X %02X, t=%X, flg=%x", p->packet.len,
			p->packet.data[0],p->packet.data[1],p->packet.data[2],p->packet.data[3],
		p->packet.beatTimeStamp, p->packet.flags);*/
	
	make_event(p->packet.data, ev);
	
    return;
}
#endif //0
// ********************************************************************
#pragma mark -

static void push_packet( Mac_Midi_Packet *packet)
{
	packet->next=NULL;
	
	if( packet_tail==NULL ){
		packet_head=packet_tail= packet;
	}else{
		packet_tail= packet_tail->next= packet;
	}
}

static Mac_Midi_Packet* mac_oms_get_nextpacket()
{
	Mac_Midi_Packet* work;
	
	work= packet_head;
	if( work==NULL ) return NULL; //no packet
	
	if( work->next==NULL ){
		packet_head= packet_tail= NULL;
	}else{
		packet_head=work->next;
	}
	
	return work;
}

void mac_oms_doevent()
{
	static int count=0;
	for(;;){
		Mac_Midi_Packet* p=mac_oms_get_nextpacket();
		if( p==NULL ) break;
		do_sequencer(p);
		count++;
		//ctl->cmsg(CMSG_ERROR, VERB_NORMAL,"got packet %d.",count++);
		free(p);
		if( count >100 ) {count=0; return;}
	}
}

static void myOMSAppHook(OMSAppHookMsg * /*packet*/, long /*appHookRefCon*/)
{
	oms_counter++;
}

static void myOMSReadHook2(OMSMIDIPacket *packet, long /*readHookRefCon*/)
{
	Mac_Midi_Packet *p;
	
	oms_counter++;
	
	p= (Mac_Midi_Packet*)malloc(sizeof(Mac_Midi_Packet));
	if( p==NULL ) return;
	p->packet= *packet;
	
	if( play_mode ){
		double past_time = get_current_calender_time() - start_time;
		p->mactime = (int32)(past_time * play_mode->rate);
	}else{
		p->mactime=0;
	}

	push_packet(p);
}

static void myTimerMessage(OMSTimerMessage* timerMessage)
{
	
}

static void mac_oms_setupdevice()
{
	OMSNodeInfoListH	h;
	OMSErr			err;
	short			portRefNum;

	err= OMSAddPort( MAC_SIGNATURE, inID, omsPortTypeInput2,
		NewOMSReadHook2(myOMSReadHook2),  0, &portRefNum);
	if( err ){
		SysBeep(0); return;
	}
	
	h= OMSGetNodeInfo(omsIncludeInputs|omsIncludeReal|omsIncludeVirtual|omsIncludeSync);
	OMSLockHandle(h);
	if( (**h).numNodes>=1 )
		ctl->cmsg(CMSG_ERROR, VERB_NORMAL,"Get success: num=%d %s",
			(**h).numNodes, p2cstr((**h).info[0].name));
	else{
		ctl->cmsg(CMSG_ERROR, VERB_NORMAL,"mac_oms_setupdevice: can't open input device.");
	}
	connects.nodeUniqueID= (**h).info[0].uniqueID;
	connects.appRefCon= 0;
	err= OMSOpenConnections(MAC_SIGNATURE, inID, 1, &connects, true);
	if( err ) ctl->cmsg(CMSG_ERROR, VERB_NORMAL,"mac_oms_setupdevice: fail OMSOpenConnections.");
	OMSDisposeHandle(h);
}

void mac_oms_setup()
{
	short	mode;
	long	ver;
	OMSErr	err;
	
	packet_head= packet_tail= NULL;
	ver= OMSVersion();
	if( ver<0x02000000 ){
		SysBeep(0); return;
	}
	
	err= OMSSignIn('TIMI', 0, "\pTiMidity++",
		NewOMSAppHook(myOMSAppHook), &mode);
	if( err ){
		SysBeep(0); return;
	}
	
	OMSTimerVersion();
	err= OMSTimerRegisterClient(&timer_client,0);
	if( err || timer_client==NULL ){
		SysBeep(0); return;
	}
	
	//err= OMSTimerSetTime(0, 4);
	//if( err ){ SysBeep(0); return; }

	err= OMSTimerSetMessageProc(timer_client, 8,
		NewOMSTimerMessageProc(myTimerMessage) );
	
	//OMSOpenCurrentStudioSetup();
	OMSMIDISetupDialog();
	mac_oms_setupdevice();
	
	err= OMSTimerTrigger();
	if( err ){ SysBeep(0); return; }

	server_setup();
	mac_osm_enable= 1;
}

void mac_oms_quit()
{
	OMSCloseConnections(MAC_SIGNATURE, inID, 1, &connects );
	OMSSignOut(MAC_SIGNATURE);
}

#endif //MAC_OMS_SERVER
