/* 
    TiMidity++ -- MIDI to WAVE converter and player
    Copyright (C) 1999 Masanao Izumo <mo@goice.co.jp>
    Copyright (C) 1995 Tuukka Toivonen <tt@cgs.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	Macintosh interface for TiMidity
	by T.Nogami	<t-nogami@happy.email.ne.jp>
		
    mac_skin.c
    Macintosh skin
*/

#ifndef TARGET_API_MAC_CARBON

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */
#include <string.h>
#include <math.h>
#include <Threads.h>

#include "timidity.h"
#include "common.h"
#include "controls.h"
#include "output.h"

#include "mac_main.h"
#include "mac_c.h"
#include "mac_skin.h"
#include "xskin.h"
#include "mac_loadBMP.h"
// *****************************************************
#pragma mark -

static int text_posx[] = {
/*     !   "   #   $   %   &   '   (   )   *   +   ,   -   .   /  */
  30, 17, 26, 30, 29, 26, 25, 16, 13, 14,  4, 19, 10, 15, 27, 21,
/* 0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ?  */ 
   0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 12, 12, 30, 28, 30,  3,
  27,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
  15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 22, 20, 23, 24, 18
};
static int text_posy[] = {
   0,  1,  0,  1,  1,  1,  1,  1,  1,  1,  2,  1,  1,  1,  1,  1,
   1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  1,  0,  2,
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  1,  1,  1,  1
};

SKIN_BMP	skins[SKIN_NUM]={
	{"main.bmp", 0, 0, {0,0,116,275} },
	{"titlebar.bmp", 0, 0, {0,0,87,334} },
	{"playpaus.bmp", 0, 0, {0,0,9,42} },
	{"numbers.bmp", 0, 0, {0,0,13,99} },
	{"cbuttons.bmp", 0, 0, {0,0,36,136} },
	{"monoster.bmp", 0, 0, {0,0,24,56} },
	{"posbar.bmp", 0, 0, {0,0,10,307} },
	{"shufrep.bmp", 0, 0, {0,0,85,95} },
	{"text.bmp", 0, 0, {0,0,18,155} },
	{"volume.bmp", 0, 0, {0,0,433,68} },
	{"balance.bmp", 0, 0, {0,0,433,68} }
};

GWorldPtr	skinWorld;
#define	SKIN_PIX	(GetPortPixMap(skinWorld))
static const RGBColor	black={0,0,0},white={0xFFFF,0xFFFF,0xFFFF},
						yellow={0xFFFF,0xFFFF,0};
static int		bmpfile_changed_flag=0;

static int open_SkinWin();
static void click_SkinWin(Point local, short modifiers);
static void update_SkinWin();
static void goaway_SkinWin(MacWindow*);
static int	message_SkinWin(int message, long param);
static Rect	panelrect={0,0,116,275};
#define		SKIN_VISCOLOR_NUM	23
static RGBColor	skin_viscolor[SKIN_VISCOLOR_NUM];

enum{ NON=0, ANALIZER=1, SCOPE=2};
static int	skin_fspe=NON;

static int isspeanaenabled;
static uint8 speana_buf[SPE_W];
static double exp_hz_table[SPE_W+1];
static int xskin_ready = 0;

#define FFTSIZE 1024 /* same as "soudspec.c" */
#define NCOLOR  32768   /* same as "soudspec.c" */
#define DEFAULT_ZOOM (44100.0/1024.0*4.0) /* tekito---- */

// *****************************************************
#pragma mark -

#define win mac_SkinWindow
MacWindow win={
	0,	//WindowRef
	open_SkinWin,
	click_SkinWin,
	update_SkinWin,
	goaway_SkinWin,
	close_default,
	message_SkinWin,
	0, 10,50,275,116,0
};


#pragma mark ============button table

SKIN_BUTTON skin_buttons[]={
	{{88,16,106,39},{{0,0},{18,0}},skin_bmp_cbuttons,-1},//prev
	{{88,39,106,62},{{0,23},{18,23}},skin_bmp_cbuttons,-1},//play
	{{88,62,106,85},{{0,46},{18,46}},skin_bmp_cbuttons,-1},//pause
	{{88,85,106,108},{{0,69},{18,69}},skin_bmp_cbuttons,-1},//stop
	{{88,108,106,130},{{0,92},{18,92}},skin_bmp_cbuttons,-1},//next
	{{89,137,105,159},{{0,114},{16,114}},skin_bmp_cbuttons,-1},//eject

	{{89,164,104,211},{{0,28},{15,28},{30,28},{45,28}},skin_bmp_shufrep,0},//shuffle
	{{89,210,104,238},{{0,0},{15,0},{30,0},{45,0}},skin_bmp_shufrep,0},//repeat

	{{PLIST_DY,PLIST_DX,PLIST_DY+PLIST_H,PLIST_DX+PLIST_W},
		{{PLIST_SY(OFF),PLIST_SX(OFF)},{PLIST_SY(OFFON),PLIST_SX(OFFON)},
			{PLIST_SY(ON),PLIST_SX(ON)},{PLIST_SY(ONOFF),PLIST_SX(ONOFF)}},skin_bmp_shufrep,0}//play list
};
#define SKIN_BTN_NUM 9

static char local_buf[300];

static int fshuf,frep,fequ,fpll;
static int fplay,fpause;
static int fremain;
static int play_val, vol_val;
static char last_text[1024];
static int last_current_time;
static int total_time;

// *****************************************************
#pragma mark -
#define SRC_PIX(pixnum)	(GetPortPixMap(skins[pixnum].gworld))

static void copy_bmp(int srcbmp, Rect src, Rect dest)
{
	GDHandle	oldGD;
	GWorldPtr	oldGW;
	//Rect		destrect;
	
	/*destrect.top=destpoint.v;
	destrect.left=destpoint.h;
	destrect.bottom=destpoint.v+ srcrect.bottom-srcrect.top;
	destrect.right=destpoint.h+ srcrect.right-srcrect.left;	*/
	
	LockPixels(SKIN_PIX);
	LockPixels(SRC_PIX(srcbmp));	
	GetGWorld(&oldGW, &oldGD);

		SetPortWindowPort(win.ref);
		RGBBackColor(&white);
		RGBForeColor(&black);
		CopyBits((BitMap*)&SRC_PIX(srcbmp),(BitMap*)&SKIN_PIX,
				&src, &dest, srcCopy,0);

	SetGWorld(oldGW, oldGD);
	UnlockPixels(SKIN_PIX);	
	UnlockPixels(SRC_PIX(srcbmp));	
}

static void skin_redisp(Rect rect)
{
	GDHandle	oldGD;
	GWorldPtr	oldGW;
	
	LockPixels(SKIN_PIX);
	GetGWorld(&oldGW, &oldGD);

		SetPortWindowPort(win.ref);
		RGBBackColor(&white);
		RGBForeColor(&black);
		CopyBits((BitMap*)&SKIN_PIX, &win.ref->portBits,
				&rect, &rect, srcCopy,0);

	SetGWorld(oldGW, oldGD);
	UnlockPixels(SKIN_PIX);
}

static void skin_remake_part(int bmpnum, int x1, int y1,
							int width, int height, int x, int y)
{
	Point	p;
	Rect	src, dest;
	
	if( skins[bmpnum].valid ){
		p.v=y; p.h=x;
		src.top=y1; src.left=x1; src.bottom=y1+height; src.right=x1+width;
		dest.top=y; dest.left=x; dest.bottom=y+height; dest.right=x+width;
		copy_bmp( bmpnum, src, dest);
		skin_redisp(dest);
	}
}

// *****************************************************
#pragma mark -


//dummy
#define True true
static int xskin_d,xskin_w,xskin_gc;
Pixmap	xskin_back=skin_bmp_main,
		xskin_titlebar=skin_bmp_titlebar,
		xskin_playpaus=skin_bmp_playpaus,
		xskin_numbers=skin_bmp_number,
		xskin_cbuttons=skin_bmp_cbuttons,
		xskin_monoster=skin_bmp_monoster,
		xskin_posbar=skin_bmp_posbar,
		xskin_shufrep=skin_bmp_shufrep,
		xskin_text=skin_bmp_text,
		xskin_volume=skin_bmp_volume,
		xskin_balance=skin_bmp_balance;


//emuration
static void XCopyArea(Display /**/, Pixmap  pix, Window /**/, GC,
		int src_x, int src_y, int width, int height, int dest_x, int dest_y)
{
	skin_remake_part( pix, src_x, src_y, width, height, dest_x, dest_y);
}

static void XSync(Display,Boolean){} //do nothing

void ts_puttext( int x0, int y0, char *message ) {
  int i,l;
  int c;
  int x,px,py;

  if ( x0 == MESSAGE_X ) {
    px=text_posx[0]*TEXT_W;
    py=text_posy[0]*TEXT_H;
    for ( i=0 ; i<31 ; i++ ) {
      x = x0+i*TEXT_W;
      XCopyArea( xskin_d, xskin_text, xskin_w, xskin_gc,
		 px, py, TEXT_W, TEXT_H, x, y0 );
    }
  } else if ( x0 == BITRATE_X ) {
    XCopyArea( xskin_d, xskin_back, xskin_w, xskin_gc,
	       111, 43, 15, 6, 111, 43 );
  } else if ( x0 == SAMPLE_X ) {
    XCopyArea( xskin_d, xskin_back, xskin_w, xskin_gc,
	       156, 43, 10, 6, 156, 43 );
  }

  l = strlen( message );
  if ( l<=0 ) return;
  for ( i=0 ; i<l ; i++ ) {
    c = (int)(message[i]);
    if ( (c>='a') && (c<='z') ) c = c-'a'+'A';
    if ( c<' ' ) c = '.';
    if ( c>'_' ) c = '.';

    c-=' ';
    if ( c>=64 ) c=0;
    px = text_posx[c]*TEXT_W;
    py = text_posy[c]*TEXT_H;
    x = x0+i*TEXT_W;
    if (( x0 == MESSAGE_X && i<31 ) ||
	( x0 == BITRATE_X && i<3  ) ||
	( x0 == SAMPLE_X  && i<2  )) {
      XCopyArea( xskin_d, xskin_text, xskin_w, xskin_gc,
		 px, py, TEXT_W, TEXT_H, x, y0 );
    }
  }
  XSync( xskin_d, True ); /* discards any events in the queue */

  if ( x0 == MESSAGE_X )
    strcpy( last_text, message );

  return;
}

/* numbers */

void ts_putnum( int x, int y, int val ) {

  int x0,y0;

  if ( (val>9) || (val<0 ) ) return;
  x0=val*9;
  y0=0;

  XCopyArea( xskin_d, xskin_numbers, xskin_w, xskin_gc,
	     x0, y0, NUM_W, NUM_H, x, y );

  return ;
}

/* cbuttons */

void ts_prev( int i ) {

  XCopyArea( xskin_d, xskin_cbuttons, xskin_w, xskin_gc,
	     PREV_SX(i), PREV_SY(i), PREV_W, PREV_H, PREV_DX, PREV_DY );

  return;
}

void ts_play( int i ) {

  XCopyArea( xskin_d, xskin_cbuttons, xskin_w, xskin_gc,
	     PLAY_SX(i), PLAY_SY(i), PLAY_W, PLAY_H, PLAY_DX, PLAY_DY );

  return;
}

void ts_pause( int i ) {

  XCopyArea( xskin_d, xskin_cbuttons, xskin_w, xskin_gc,
	     PAUSE_SX(i), PAUSE_SY(i), PAUSE_W, PAUSE_H, PAUSE_DX, PAUSE_DY);

  return;
}

void ts_stop ( int i ) {

  XCopyArea( xskin_d, xskin_cbuttons, xskin_w, xskin_gc,
	     STOP_SX(i), STOP_SY(i), STOP_W, STOP_H, STOP_DX, STOP_DY );

  return;
}

void ts_next( int i ) {

  XCopyArea( xskin_d, xskin_cbuttons, xskin_w, xskin_gc,
	     NEXT_SX(i), NEXT_SY(i), NEXT_W, NEXT_H, NEXT_DX, NEXT_DY );

  return;
}

void ts_eject( int i ) {

  XCopyArea( xskin_d, xskin_cbuttons, xskin_w, xskin_gc,
	     EJECT_SX(i), EJECT_SY(i), EJECT_W, EJECT_H, EJECT_DX, EJECT_DY );
  return;
}

/* titlebar */

void ts_titlebar( int i ) {

  XCopyArea( xskin_d, xskin_titlebar, xskin_w, xskin_gc,
	     TITLEBAR_SX(i), TITLEBAR_SY(i),
	     TITLEBAR_W, TITLEBAR_H, TITLEBAR_DX, TITLEBAR_DY );
  return;
}

void ts_exitbutton( int i ) {
  
  XCopyArea( xskin_d, xskin_titlebar, xskin_w, xskin_gc,
	     EXITBUTTON_SX(i), EXITBUTTON_SY(i),
	     EXITBUTTON_W, EXITBUTTON_H, EXITBUTTON_DX, EXITBUTTON_DY );

  return;
}

void ts_menubutton( int i ) {

  XCopyArea( xskin_d, xskin_titlebar, xskin_w, xskin_gc,
	     MENUBUTTON_SX(i), MENUBUTTON_SY(i),
	     MENUBUTTON_W, MENUBUTTON_H, MENUBUTTON_DX, MENUBUTTON_DY );
  return;
}

void ts_iconbutton( int i ) {

  XCopyArea( xskin_d, xskin_titlebar, xskin_w, xskin_gc,
	     ICONBUTTON_SX(i), ICONBUTTON_SY(i),
	     ICONBUTTON_W, ICONBUTTON_H, ICONBUTTON_DX, ICONBUTTON_DY );
  return;
}

void ts_minibutton( int i ) {

  XCopyArea( xskin_d, xskin_titlebar, xskin_w, xskin_gc,
	     MINIBUTTON_SX(i), MINIBUTTON_SY(i),
	     MINIBUTTON_W, MINIBUTTON_H, MINIBUTTON_DX, MINIBUTTON_DY );
  return;
}

/* monoster */

void ts_mono( int i ) {

  XCopyArea( xskin_d, xskin_monoster, xskin_w, xskin_gc,
	     MONO_SX(i), MONO_SY(i),
	     MONO_W, MONO_H, MONO_DX, MONO_DY );
  return;
}

void ts_stereo( int i ) {

  XCopyArea( xskin_d, xskin_monoster, xskin_w, xskin_gc,
	     STEREO_SX(i), STEREO_SY(i),
	     STEREO_W, STEREO_H, STEREO_DX, STEREO_DY );
  return;
}

/* playpaus */

static void ts_pstate( int i ) {

  XCopyArea( xskin_d, xskin_playpaus, xskin_w, xskin_gc,
	     PSTATE1_SX(i), PSTATE1_SY(i),
	     PSTATE1_W, PSTATE1_H, PSTATE1_DX, PSTATE1_DY );
  XCopyArea( xskin_d, xskin_playpaus, xskin_w, xskin_gc,
	     PSTATE2_SX(i), PSTATE2_SY(i),
	     PSTATE2_W, PSTATE2_H, PSTATE2_DX, PSTATE2_DY );
  return;
}

/* shufrep */

void ts_shuf( int i ) {

  XCopyArea( xskin_d, xskin_shufrep, xskin_w, xskin_gc,
	     SHUF_SX(i), SHUF_SY(i),
	     SHUF_W, SHUF_H, SHUF_DX, SHUF_DY );

  return;
}

void ts_rep( int i ) {

  XCopyArea( xskin_d, xskin_shufrep, xskin_w, xskin_gc,
	     REP_SX(i), REP_SY(i),
	     REP_W, REP_H, REP_DX, REP_DY );

  return;
}

void ts_equ( int i ) {

  XCopyArea( xskin_d, xskin_shufrep, xskin_w, xskin_gc,
	     EQU_SX(i), EQU_SY(i),
	     EQU_W, EQU_H, EQU_DX, EQU_DY );
  return;
}

void ts_plist( int i ) {

  XCopyArea( xskin_d, xskin_shufrep, xskin_w, xskin_gc,
	     PLIST_SX(i), PLIST_SY(i),
	     PLIST_W, PLIST_H, PLIST_DX, PLIST_DY );
  return;
}

/* posbar */

int ts_pos( int i, int j ) {

  int x,y;
  int p;

  if ( j<0 ) p=-j;
  else {
    if (j<POS_MIN_DX) j=POS_MIN_DX;
    if (j>POS_MAX_DX) j=POS_MAX_DX;
    p = 100*(j-POS_MIN_DX)/(POS_MAX_DX-POS_MIN_DX);
  }

  x = POS_MIN_DX + (POS_MAX_DX-POS_MIN_DX)*p/100;
  y = POS_DY;

  XCopyArea( xskin_d, xskin_posbar, xskin_w, xskin_gc,
	     BAR_SX, BAR_SY, BAR_W, BAR_H, BAR_DX, BAR_DY );
  XCopyArea( xskin_d, xskin_posbar, xskin_w, xskin_gc,
	     POS_SX(i), POS_SY(i), POS_W, POS_H, x, y );
  return p;
}

int ts_volume( int i, int j ) {

  int x,y;
  int t,p;

  if ( j<0 ) p=-j;
  else {
    if (j<VOL_MIN_DX) j=VOL_MIN_DX;
    if (j>VOL_MAX_DX) j=VOL_MAX_DX;
    p = 100*(j-VOL_MIN_DX)/(VOL_MAX_DX-VOL_MIN_DX);
  }

  x = VOL_MIN_DX + (VOL_MAX_DX-VOL_MIN_DX)*p/100;
  y = VOL_DY;
  t=27*(int)p/100;
  XCopyArea( xskin_d, xskin_volume, xskin_w, xskin_gc,
	     VOLUME_SX, t*VOLUME_H,
	     VOLUME_W, VOLUME_H-2, VOLUME_DX, VOLUME_DY );
  XCopyArea( xskin_d, xskin_volume, xskin_w, xskin_gc,
	     VOL_SX(i), VOL_SY(i), VOL_W, VOL_H, x, y );
  return p;
}

int ts_pan( int i, int j ) {

  int x,y;
  int t,p;

  if ( j<0 ) p=-j;
  else {
    if (j<PAN_MIN_DX) j=PAN_MIN_DX;
    if (j>PAN_MAX_DX) j=PAN_MAX_DX;
    p = 100*(j-PAN_MIN_DX)/(PAN_MAX_DX-PAN_MIN_DX);
  }

  x = PAN_MIN_DX + (PAN_MAX_DX-PAN_MIN_DX)*p/100;
  y = PAN_DY;
  t=27*((p>50)?((float)p-50)/50:(50-(float)p)/50);
  if ( t<2 ) t=0;
  XCopyArea( xskin_d, xskin_volume, xskin_w, xskin_gc,
	     PANPOT_SX, t*PANPOT_H,
	     PANPOT_W, PANPOT_H-2, PANPOT_DX, PANPOT_DY );
  XCopyArea( xskin_d, xskin_volume, xskin_w, xskin_gc,
	     PAN_SX(i), PAN_SY(i),
	     PAN_W, PAN_H, x, y );
  return p;
}

// *****************************************************
#pragma mark -

static const Rect rect_volume_all={57,107,69,175};

static void skin_volume(int now, int total, int grab)
{
	int		x, val;
	static int	save_total=100;
	
	if( total==-1 ) total=save_total;
	else save_total=total;
	
	if( total==0 ) total=100;
	if( now>total ) now=total;
	//x=68+107*now/total;
	val= (rect_volume_all.right-rect_volume_all.left-17)*now/total;
	x= rect_volume_all.left+val;	
	skin_remake_part(skin_bmp_volume,  0,(val/2)*15,  68, 12, 107, 57);//all bar
	skin_remake_part(skin_bmp_volume, grab?0:15, 422,  14, 11,   x, 58);//slider
	skin_redisp(rect_volume_all);
}

static void HandleVolume()
{
#define BUFSIZE 80
	int amp, width;
	Point	last={0,0}, curr;
	char	tmp[BUFSIZE];
	
	width= rect_volume_all.right- rect_volume_all.left-14;
	do{
		SetPortWindowPort(win.ref);
		GetMouse(&curr);	//get local
		if( curr.h==last.h /*&& curr.v==last.v*/ ) continue; //mouse not moved
		amp= 255* (curr.h-6 - rect_volume_all.left)/width;
		if( amp>255 ) amp=255;
		if( amp<0 ) amp=0;
		skin_volume(amp, 255, true);
		mac_setVolume(amp);
		last=curr;
		
		snprintf( tmp, BUFSIZE, " volume: %d%%", 100*amp/255 );
		ts_puttext( MESSAGE_X, MESSAGE_Y, tmp );

	}while( YieldToAnyThread(), Button() );
	skin_volume(amp, 255, false); //mouse release
}


/*#define SKIN_ACTION_PREV() {mac_rc=RC_PREVIOUS;mac_HandleControl();}
#define SKIN_ACTION_PLAY() {mac_rc=RC_CONTINUE;mac_HandleControl();}
#define SKIN_ACTION_PAUSE() {mac_rc=RC_TOGGLE_PAUSE;mac_HandleControl();}
#define SKIN_ACTION_STOP() {mac_rc=RC_QUIT;mac_HandleControl();}
#define SKIN_ACTION_NEXT() {mac_rc=RC_NEXT;mac_HandleControl();}
#define SKIN_ACTION_EJECT()
#define SKIN_ACTION_EXIT()
*/

#define SKIN_ACTION_SHUFFLE(onoff)	(gShuffle=(onoff))
#define SKIN_ACTION_REPEAT(onoff)	(skin_f_repeat=(onoff))

#define SKIN_ACTION_PLAYLIST()  SHOW_WINDOW(mac_ListWindow)

// *****************************************************
#pragma mark -

#define ISIN(x,y,x0,y0,w,h) ( (x>=x0)&&(x<x0+w)&&(y>=y0)&&(y<y0+h) )?1:0

static int skin_mouseDown(int x, int y)
{
	int		last_window_x=x,
			last_window_y=y,
			pr= -1;
			

	if ( ISIN( x, y,EXITBUTTON_DX,EXITBUTTON_DY,
			   EXITBUTTON_W,EXITBUTTON_H ) ) {
		ts_exitbutton(ON);pr=TS_EXITBUTTON;

	} else if ( ISIN( x, y, PREV_DX, PREV_DY, PREV_W, PREV_H ) ) {
		ts_prev(ON);pr=TS_PREV;
	} else if ( ISIN( x, y, PLAY_DX, PLAY_DY, PLAY_W, PLAY_H ) ) {
		ts_play(ON);pr=TS_PLAY;
	} else if ( ISIN( x, y, PAUSE_DX, PAUSE_DY, PAUSE_W, PAUSE_H ) ) {
		ts_pause(ON);pr=TS_PAUSE;
	} else if ( ISIN( x, y, STOP_DX, STOP_DY, STOP_W, STOP_H ) ) {
		ts_stop(ON);pr=TS_STOP;
	} else if ( ISIN( x, y, NEXT_DX, NEXT_DY, NEXT_W, NEXT_H ) ) {
		ts_next(ON);pr=TS_NEXT;
	} else if ( ISIN( x, y, EJECT_DX, EJECT_DY, EJECT_W, EJECT_H ) ) {
		ts_eject(ON);pr=TS_EJECT;
	} else if ( ISIN( x, y,164, 89, 47, 15 ) ) {  /* shuffle */
		if ( fshuf==0 ) {
			ts_shuf(OFFON);pr=TS_SHUFON;
		} else { 
			ts_shuf(ONOFF);pr=TS_SHUFOFF;
		}
	} else if ( ISIN( x, y,210, 89, 28, 15 ) ) {  /* repeat */
		if ( frep==0 ) {
			ts_rep(OFFON);pr=TS_REPON;
		} else {
			ts_rep(ONOFF);pr=TS_REPOFF;
		}
	} else if ( ISIN( x, y,219, 58, 23, 12 ) ) {  /* equalizer */
		if ( fequ==0 ) {
			ts_equ(OFFON);pr=TS_EQUON;
		} else {
			ts_equ(ONOFF);pr=TS_EQUOFF;
		}
	} else if ( ISIN( x, y,242, 58, 23, 12 ) ) {  /* playlist */
		if ( fpll==0 ) {
			ts_plist(OFFON);pr=TS_PLISTON;
		} else {
			ts_plist(ONOFF);pr=TS_PLISTOFF;
		}

	} else if ( ISIN( x, y, MENUBUTTON_DX, MENUBUTTON_DY,
					  MENUBUTTON_W, MENUBUTTON_H ) ) {
		ts_menubutton(ON);pr=TS_MENUBUTTON;
	} else if ( ISIN( x, y, ICONBUTTON_DX, ICONBUTTON_DY,
					  ICONBUTTON_W, ICONBUTTON_H ) ) {
		ts_iconbutton(ON);pr=TS_ICONBUTTON;
	} else if ( ISIN( x, y, MINIBUTTON_DX, MINIBUTTON_DY,
					  MINIBUTTON_W, MINIBUTTON_H ) ) {
		ts_minibutton(ON);pr=TS_MINIBUTTON;

		/*
		  }	else if ( ISIN( x, y,POS_MIN_DX+(POS_MAX_DX-POS_MIN_DX)*play_val/100,
		  POS_DY, POS_W, POS_H ) ) {
		  ts_pos( ON, -play_val );pr=TS_POS;
		  */
		//} else if ( ISIN( x, y,VOL_MIN_DX+(VOL_MAX_DX-VOL_MIN_DX)*vol_val/100,
	} else if ( ISIN( x, y, VOL_MIN_DX, VOL_DY,
					  VOL_MAX_DX-VOL_MIN_DX+VOL_W, VOL_H ) ) {
		HandleVolume();
		//ts_volume( ON, -vol_val );pr=TS_VOLUME;
		//sprintf( tmp, " volume: %d%%", vol_val );
		//ts_puttext( MESSAGE_X, MESSAGE_Y, tmp );
		//last_puttext_time=last_current_time;
		
		/*
		  } else if ( ISIN( x, y,PAN_MIN_DX+(PAN_MAX_DX-PAN_MIN_DX)*pan_val/100,
		  PAN_DY, PAN_W, PAN_H ) ) {
		  ts_pan( ON, -pan_val );pr=TS_PAN;
		  */
		
	} else if ( ISIN( x, y, MIN_H_X, MIN_H_Y,
					  SEC_L_X+NUM_W-MIN_H_X, NUM_H ) ) {
		int min,sec;
		fremain=(fremain==0)?1:0;
		sec=(fremain==0)?last_current_time:total_time-last_current_time;
		min =sec/60;
		sec-=min*60;
		ts_putnum( MIN_H_X, MIN_H_Y, min/10 );
		ts_putnum( MIN_L_X, MIN_L_Y, min%10 );
		ts_putnum( SEC_H_X, SEC_H_Y, sec/10 );
		ts_putnum( SEC_L_X, SEC_L_Y, sec%10 );
		
	} else if ( ISIN( x, y, SPE_SX, SPE_SY, SPE_W, SPE_H ) ) {
		pr=TS_SPECTRUM;
		skin_fspe = (skin_fspe+1)%3;
	} else {
		//XRaiseWindow( xskin_d, xskin_w );
	}
	
	return pr;
}

static void skin_mouseUp(int pr)
{
	int		last_window_x = -1,
			last_window_y = -1;
	
	switch( pr ) {
		
	case TS_EXITBUTTON:
		ts_exitbutton(OFF);
		//xskin_pipe_write("Q");
		SKIN_ACTION_EXIT();
		break;
		
	case TS_PREV:
		ts_prev(OFF);
		//xskin_pipe_write("B");
		SKIN_ACTION_PREV();
		break;
	case TS_PLAY:
		//xskin_pipe_write("P");
		SKIN_ACTION_PLAY();
		fplay=1;fpause=0;
		ts_play(OFF);ts_pause(OFF);
		ts_pstate( PSTATE_PLAY );
		break;
	case TS_PAUSE:
		ts_pause(OFF);
		//if ( fplay ==1 ) {
		//	if ( fpause==0 ) {
				ts_pstate( PSTATE_PAUSE );
				fpause=1;
		//	} else {
				ts_pstate( PSTATE_PLAY );
				fpause=0;
		//	}
			//xskin_pipe_write("U");
			SKIN_ACTION_PAUSE();
		//}
		break;
	case TS_STOP:
		fpause=0;fplay=0;
		ts_pause(OFF);ts_play(OFF);ts_stop(OFF);
		ts_pstate( PSTATE_STOP );
		//xskin_pipe_write("S");
		SKIN_ACTION_STOP();
		break;
	case TS_NEXT:
		ts_next(OFF);
		//xskin_pipe_write("N");
		SKIN_ACTION_NEXT();
		break;
	case TS_EJECT:
		ts_eject(OFF);
		SKIN_ACTION_EJECT();
		break;
		
	case TS_SHUFON:
		ts_shuf(ON);fshuf=1;
		fplay=1;fpause=0;
		ts_pstate( PSTATE_PLAY );
		//xskin_pipe_write("D 1");
		SKIN_ACTION_SHUFFLE(ON);
		break;
	case TS_SHUFOFF:
		ts_shuf(OFF);fshuf=0;
		fplay=0;fpause=0;
		ts_pstate( PSTATE_STOP );
		//xskin_pipe_write("D 2");
		SKIN_ACTION_SHUFFLE(OFF);
		break;
	case TS_REPON:
		ts_rep(ON);frep=1;
		//xskin_pipe_write("R 1");
		SKIN_ACTION_REPEAT(ON);
		break;
	case TS_REPOFF:
		ts_rep(OFF);frep=0;
		//xskin_pipe_write("R 0");
		SKIN_ACTION_REPEAT(OFF);
		break;
		
	case TS_EQUON:
		ts_equ(ON);fequ=1;break;
	case TS_EQUOFF:
		ts_equ(OFF);fequ=0;break;
		
	case TS_PLISTON:
		ts_plist(ON);fpll=1; SKIN_ACTION_PLAYLIST(); break;
	case TS_PLISTOFF:
		ts_plist(OFF);fpll=0;break;
		
	case TS_MENUBUTTON:
		ts_menubutton(OFF);break;
	case TS_ICONBUTTON:
		ts_iconbutton(OFF);break;
	case TS_MINIBUTTON:
		ts_minibutton(OFF);break;
		
		/*
		  case TS_POS:
		  ts_pos( OFF, -play_val );break;
		  */
	case TS_VOLUME:
		ts_volume( OFF, -vol_val );break;
		/*
		  case TS_PAN:
		  ts_pan( OFF, -pan_val );break;
		  */
		
	case TS_SPECTRUM:
		//skin_fspe = (skin_fspe+1)%3;
		skin_need_speana= !skin_need_speana;
		//mac_rc= RC_TOGGLE_CTL_SPEANA;
		break;
		
	default:
		break;
	}
	pr=-1;
}

static pascal void *xskin_jobs( void *params )
{
	Point	local= *(Point*)params;
	int	what=mouseDown, x=local.h, y=local.v;
    //int window_x,window_y;
    static int pr=-1;
    int z;
    int p;
    int master_volume=0;
    char file_name[1024], tmp[1024];
    int last_puttext_time;
    int last_window_x=-1, last_window_y=-1;
    //int max_files;
    int i;
    //static struct timeval tv;
    //Window t_w;
    //unsigned int t_width, t_height, t_border, t_depth;
    
    z=1;
    last_puttext_time=0;
    last_current_time=0;
    
	while( z ) {
	    
		XFlush( xskin_d );
		
		if ( i!=0 ) {
			//xskin_pipe_read( local_buf, sizeof(local_buf) );
			switch (local_buf[0]) {
				
			case 'A': /* total time */
				total_time=atoi( local_buf+2 );
				break;
				
			case 'T': /* current time */
				{
					int min,sec;
					sscanf( local_buf+2, "%02d:%02d", &min, &sec );
					i=min*60+sec;
					if ( fremain==1 ) {
						sec =total_time-i;
						min =sec/60;
						sec-=min*60;
					}
					if ( i != last_current_time ) {
						ts_putnum( MIN_H_X, MIN_H_Y, min/10 );
						ts_putnum( MIN_L_X, MIN_L_Y, min%10 );
						ts_putnum( SEC_H_X, SEC_H_Y, sec/10 );
						ts_putnum( SEC_L_X, SEC_L_Y, sec%10 );
						p=100*i/total_time;
						play_val=ts_pos( OFF, -p );
						last_current_time=i;
						
						if ( last_current_time - last_puttext_time == 3 ) {
							                                  /* 3 sec */
							sprintf( tmp, "%s [%02d:%02d]",
									 file_name, total_time/60, total_time%60 );
							ts_puttext( MESSAGE_X, MESSAGE_Y, tmp );
						}
					}
				}
			break;
			
			case 'L': /* lylics/message */
				ts_puttext( MESSAGE_X, MESSAGE_Y, local_buf+2 );
				last_puttext_time=last_current_time;
				break;
				
			case 'F': /* filename */
				strncpy( file_name, local_buf+2, 1023 );
				file_name[1023]=0;
				break;
				
			case 'O': /* off the play button */
				fplay=0;
				ts_play(OFF);
				break;
				
			case 'V': /* master volume */
				master_volume=atoi( local_buf+2 );
				p=100*((master_volume<200)?master_volume:200)/200;
				                                  /* max:200% */
				vol_val=ts_volume( OFF, -p );
				break;
				
			case 'Q': /* quit */
				z=1;
				break;
				
			default:
				break;
			}
		}
		
		//if ( fspe==1 )
		//ts_spectrum(fspe);
		
		//if ( XPending( xskin_d )==0 ) continue;
		//XNextEvent( xskin_d, &e );
		
		switch ( what ) {
			/*
			  case KeyPress:
			  z=0;
			  break;
			  */
			
			/*case Expose:
			  repaint();
			  break;*/
			
			/*case EnterNotify:
			  {
			  Cursor cs;
			  ts_titlebar(ON);
			  cs = XCreateFontCursor( xskin_d, XC_top_left_arrow );
			  XDefineCursor( xskin_d, xskin_w, cs );
			  }
			  break;*/
			
			/*case LeaveNotify:
			  ts_titlebar(OFF);
			  XUndefineCursor( xskin_d, xskin_w );
			  break;*/
#if 0
		case MotionNotify:
			while( XCheckMaskEvent( xskin_d, Button1MotionMask, &e ) ) {
				XNextEvent( xskin_d, &e );
			}
			x = e.xbutton.x;
			y = e.xbutton.y;
			switch( pr ) {
				
				/*
				  case TS_POS:
				  play_val=ts_pos( ON, x );break;
				  */
			case TS_VOLUME:
				vol_val=ts_volume( ON, x );
				i=master_volume;
				master_volume=200*vol_val/100;
				sprintf( local_buf, "V %d", master_volume-i );
				xskin_pipe_write( local_buf );

				sprintf( tmp, " volume: %d%%", vol_val );
				ts_puttext( MESSAGE_X, MESSAGE_Y, tmp );
				last_puttext_time=last_current_time;
				break;
				/*
				  case TS_PAN:
				  pan_val=ts_pan( ON, x );break;
				  */
	
			default:
				if ( x != last_window_x || y != last_window_y ) {
					window_x += x-last_window_x;
					window_y += y-last_window_y;
					XMoveWindow( xskin_d, xskin_w,
								 window_x, window_y );
				}
				break;
			}
			break;
#endif

		case mouseDown:
			pr= skin_mouseDown(x,y);
			while(Button()){ YieldToAnyThread(); }
			what= mouseUp;
			break;
			
		case mouseUp:
			skin_mouseUp(pr);
			z=0;
			break;
			
		default:
			break;
		}
	} //end while
	
	return 0;
}


// *****************************************************
#pragma mark -

static void clear_gworld(GWorldPtr gworld)
{
	GDHandle	oldGD;
	GWorldPtr	oldGW;
	//Rect		destrect;
	
	/*destrect.top=destpoint.v;
	destrect.left=destpoint.h;
	destrect.bottom=destpoint.v+ srcrect.bottom-srcrect.top;
	destrect.right=destpoint.h+ srcrect.right-srcrect.left;	*/
	
	LockPixels(GetPortPixMap(gworld));
	GetGWorld(&oldGW, &oldGD);
	SetGWorld(gworld,0);

		PaintRect(&gworld->portRect);

	SetGWorld(oldGW, oldGD);
	UnlockPixels(SKIN_PIX);	
}

static OSErr skin_setup()
{
	OSErr	err;
	int		i;
	Rect	rect={0,0,450,400};
	
	err=NewGWorld(&skinWorld, 16, &rect,0,0,keepLocal);
	if( err ) return err;
	clear_gworld(skinWorld);
	
	for( i=0; i<SKIN_NUM; i++){
		err=NewGWorld(&skins[i].gworld, 16, &rect,0,0,keepLocal);
		if( err ) return err;
	}
	skin_buttons[skin_btn_shuffle].led=gShuffle;
	skin_buttons[skin_btn_repeat].led=skin_f_repeat;
	local_buf[0]=0;
	return noErr;
}

// *****************************************************
#pragma mark -

/*static void ts_titlebar( int i )
{
	skin_remake_part(  skin_bmp_titlebar,
		TITLEBAR_SX(i), TITLEBAR_SY(i),
		TITLEBAR_W, TITLEBAR_H, TITLEBAR_DX, TITLEBAR_DY );
	return;
}*/


static void skin_update_button(SKIN_BUTTON *b, int state)
{
	int		width=b->rect.right- b->rect.left,
			height=b->rect.bottom- b->rect.top;
	Point	p;
	
	//if( b->led==BTN_LED_ON ) state+=2;
	//p=b->image[state];
	p=b->image[state+(b->led==BTN_LED_ON?2:0)];
	skin_remake_part(b->bmpnum, p.h, p.v, width, height,  b->rect.left, b->rect.top);
	if( state==1 && b->led!=BTN_LED_NON ) b->led= !b->led; //toggle
}

static void PutText( int x0, int y0, const char *text )
{
	int i,l;
	int c;
	int x,px,py;

	l = strlen( text );
	for ( i=0 ; i<l ; i++ ) {
		c = (int)text[i];
		if ( (c>='a') && (c<='z') ) c = c-'a'+'A';
		if ( c<' ' ) c = ' ';
		if ( c>'_' ) c = ' ';

		c-=' ';
		if ( c>=64 ) c=0;
		px = text_posx[c]*5;
		py = text_posy[c]*6;
		x = x0+i*5;
		skin_remake_part(skin_bmp_text,px, py, 5, 6, x, y0 );
	}
}

static void skin_put_time(int min, int sec)
{
	Rect	r={26,48,39,99};
	
	if( !win.show ) return;
	if( min>59 ) min=59;
	if( min<0  ) min=0;
	if( sec>59 ) sec=59;
	if( sec<0 ) sec=0;
	skin_remake_part(skin_bmp_number, (min/10)*9,  0,   9, 13,  48, 26);//sec_l
	skin_remake_part(skin_bmp_number,  min%10*9,  0,   9, 13,  60, 26);//min_l
	skin_remake_part(skin_bmp_number, (sec/10)*9,  0,   9, 13,  78, 26);//sec_h
	skin_remake_part(skin_bmp_number, sec%10*9,  0,   9, 13,  90, 26);//sec_l
	skin_redisp(r);
}

static void skin_filename(const char str[])
{
	Rect	r={27,112,33,269};
	
	if( !win.show ) return;
	//PutText( 112, 27, str );
	ts_puttext( MESSAGE_X, MESSAGE_Y, str );
	skin_redisp(r);

	/*filename event means start of play, initialize panel*/
	{
		char	buf[10];
		snprintf(buf, 10,"%2d",play_mode->rate/1000); //as kHz
		PutText( 111, 43, "---" ); /* bit-rate */
		PutText( 156, 43, buf  ); /* sample-rate */
	
		if ( (play_mode->encoding & PE_MONO)==0 ) {
			ts_mono(OFF);
			ts_stereo(ON);
  		} else {
			ts_mono(ON);
			ts_stereo(OFF);
		}
	}
	
}

static void skin_posbar(int now, int total)
{
	Rect	all={72,16,82,264};
	int		x;
	static int	save_total=100;
	
	if( total==-1 ) total=save_total;
	else save_total=total;
	
	if( !win.show ) return;
	if( total==0 ) total=100;
	if( now>total ) now=total;
	x=16+219*now/total;
	skin_remake_part(skin_bmp_posbar,  0,  0, 248, 10,  16, 72);//all bar
	skin_remake_part(skin_bmp_posbar,248,  0,  29, 10,   x, 72);//slider
	skin_redisp(all);
}

static void skin_balance(int now, int total)
{
	Rect	all={57,178,69,215};
	int		x, bmpnum;
	static int	save_total=100;
	
	if( total==-1 ) total=save_total;
	else save_total=total;
	
	if( total==0 ) total=100;
	if( now>total ) now=total;
	x=168+47*now/total;
	bmpnum= ( skins[skin_bmp_balance].valid? 
	              skin_bmp_balance:skin_bmp_volume );
	skin_remake_part(bmpnum,  10,  0,  37, 13, 178, 57);//all bar
	skin_remake_part(bmpnum, 15,422,  14, 11,   x, 58);//slider
	//skin_remake_part(skin_bmp_volume, 16,422,  14, 11,   x, 58);//slider
	skin_redisp(all);
}

static void skin_remake_disp(Rect /*rect*/)
{
	int i;
	
	skin_buttons[skin_btn_repeat].led= skin_f_repeat;
	skin_buttons[skin_btn_shuffle].led= gShuffle;

	skin_remake_part(skin_bmp_main,      0,  0, 275,116,   0,  0);//main
	ts_titlebar(ON);
	//skin_remake_part(skin_bmp_cbuttons,  0,  0,  23, 16,  16, 88);//prev
	//skin_remake_part(skin_bmp_cbuttons, 23,  0,  23, 16,  39, 88);//play
	//skin_update_button(skin_buttons[skin_btn_prev], 0);
	//skin_update_button(skin_buttons[skin_btn_play], 0);
	for( i=0; i<SKIN_BTN_NUM; i++){
		skin_update_button(&skin_buttons[i], 0);
	}
	//skin_remake_part(skin_bmp_cbuttons, 46,  0,  23, 16,  62, 88);//pause
	//skin_remake_part(skin_bmp_cbuttons, 69,  0,  23, 16,  85, 88);//stop
	//skin_remake_part(skin_bmp_cbuttons, 92,  0,  23, 16, 108, 88);//next
	//skin_remake_part(skin_bmp_cbuttons,114,  0,  21, 16, 136, 88);//eject

	skin_remake_part(skin_bmp_playpaus,  0,  0,   9,  9,  26, 28);//play/pause
	skin_remake_part(skin_bmp_playpaus, 27,  0,   3,  9,  24, 28);//play/pause

	//skin_remake_part(skin_bmp_shufrep, 28,  0,  47, 15, 164, 89);//shuffle
	//skin_remake_part(skin_bmp_shufrep,  0,  0,  28, 15, 210, 89);//repeat
	skin_remake_part(skin_bmp_shufrep,  0, 61,  23, 12, 219, 58);//equalizer
	skin_remake_part(skin_bmp_shufrep, 23, 61,  23, 12, 242, 58);//playlist
	
	skin_remake_part(skin_bmp_monoster,  29,  0,  27, 12, 212, 41);//mono
	//skin_remake_part(skin_bmp_monoster,  0,  0,  27, 12, 239, 41);//stereo

	skin_posbar(0, -1);

	skin_put_time(12, 34);
	
	skin_volume(mac_amplitude, 255, false);
	skin_balance(50, 100);
	//skin_remake_part(skin_bmp_volume, 16,423,  13, 10, 107, 59);//volume
	//skin_remake_part(skin_bmp_volume, 16,423,  13, 10, 168, 59);//volume 

	//PutText( 111, 43, "128" ); /* bit-rate */
	//PutText( 156, 43, "44"  ); /* sample-rate */
	//PutText( 109, 24, "This is Test Texts."  );   /* text */
}

/*static int strcasecmp_tail(const char longstr[], const char tailstr[])
{
	int	longlen=strlen(longstr);
	int	taillen=strlen(tailstr);
	
	if( longlen<taillen ) return -1;
	return strcasecmp(&longstr[longlen-taillen], tailstr);
}*/

static int filename2id(const char filename[])
{
	int i;
	for( i=0; i<SKIN_NUM; i++ ){
		if( strtailcasecmp(filename, skins[i].filename)==0 ){
			return i;
		}
	}
	return -1; // no match
}

// *****************************************************
#pragma mark -

static int skin_load_bmp(const char filename[])
{
	int		bmpnum;
	OSErr	err;
	GDHandle	oldGD;
	GWorldPtr	oldGW;
	int		width, height;

	bmpnum=filename2id(filename);
	if( bmpnum==-1 ) return -1;

	GetGWorld(&oldGW, &oldGD);
	LockPixels(SRC_PIX(bmpnum));
		SetGWorld(skins[bmpnum].gworld,0);
		err= loadBMP( filename, &width, &height);
		SetGWorld(oldGW, oldGD);
	UnlockPixels(SRC_PIX(bmpnum));
	
	if( err ) return err;
	skins[bmpnum].valid=true;
	bmpfile_changed_flag=true;
	SetPortWindowPort(win.ref);	InvalWindowRect(win.ref,&panelrect);
	if( bmpnum==skin_bmp_main ) SHOW_WINDOW(mac_SkinWindow);
	//skin_remake_disp(panelrect);
	//skin_redisp(panelrect);
	return 0; //no error	
}

void	read_viscolor(const char * viscolor_file)
{
	struct	timidity_file *tf;
	int		i;
#define BUFF_SIZE 80
	char	buf[BUFF_SIZE];
	
	//ctl->cmsg(CMSG_INFO, VERB_NORMAL, "read_viscolor start");
	tf = open_file( viscolor_file, 0, OF_SILENT );
	if( tf==NULL ){
		ctl->cmsg(CMSG_INFO, VERB_NORMAL, "read_viscolor open error");
		goto finish2;
	}
	
	for(i=0; i<SKIN_VISCOLOR_NUM; i++){
		int num, red,green,blue;
		if(  tf_gets(buf, BUFF_SIZE, tf)==NULL ) goto finish1;
		num=sscanf(buf, "%d, %d, %d", &red,&green,&blue);
		if( num!=3 ){
			ctl->cmsg(CMSG_INFO, VERB_NORMAL, "read_viscolor sscanf error line:%d",i);
			goto finish1;
		}
		skin_viscolor[i].red  =red*256;
		skin_viscolor[i].green=green*256;
		skin_viscolor[i].blue =blue*256;
	}
	//ctl->cmsg(CMSG_INFO, VERB_NORMAL, "read_viscolor finished normally");

finish1:
	close_file( tf );
finish2:
	return;
}

// *****************************************************
#pragma mark -

#define SPEC_WIDTH 64
#define SPEC_HEIGHT 15

/*static void skin_draw_spectrum_normal(const int value[SPEC_WIDTH])
{
	int	i, x, y, height;
	
	for( i=0; i<SPEC_WIDTH; i++){
		height=value[i]/4096;
		if(height>SPEC_HEIGHT) height=SPEC_HEIGHT;
		RGBForeColor(&skin_viscolor[17-height]);
		x=i+30;
		MoveTo(x,58); Line(0,-height);
		
		y=58-height;
		y &= ~1;  // y: to even number
		if( x & 1 ){ //case x is odd
			RGBForeColor(&skin_viscolor[0]);
			LineTo(x,43);
		}else{ //case x is eaven
			for( ; y>=43; ){
				SetCPixel(x, y, &skin_viscolor[1]);	y--;
				SetCPixel(x, y, &skin_viscolor[0]);	y--;
			}
		}
	}	
}*/

static void skin_draw_spectrum_normal(const int value[SPEC_WIDTH])
{
	int	i, x, y, y0, height;
	
	for( i=0; i<SPEC_WIDTH; i++){
		height=value[i]/4096;
		if(height>SPEC_HEIGHT) height=SPEC_HEIGHT;
		x=i+30;
				
		y=43;
		y0=58-height;
		
		if( x & 1 ){		//case x is odd
			RGBForeColor(&skin_viscolor[0]);
			MoveTo(x,y);
			LineTo(x,y0);
		}else{ 				//case x is eaven
			for( ; y0>y; ){
				SetCPixel(x, y, &skin_viscolor[0]);	y++;
				SetCPixel(x, y, &skin_viscolor[1]);	y++;
			}
		}
		
		if( height ){
			RGBForeColor(&skin_viscolor[17-height]);
			MoveTo(x,y0);
			LineTo(x,58);
		}
	}	
}

static void skin_draw_spectrum_fire(const int value[SPEC_WIDTH])
{
	int	i, c, x, y, y0, height;
	
	SetPortWindowPort(win.ref);
	for( i=0; i<SPEC_WIDTH; i++){
		height=value[i]/1024;
		if(height>SPEC_HEIGHT) height=SPEC_HEIGHT;
		x=i+30;
				
		y=43;
		y0=58-height;
		
		if( x & 1 ){		//case x is odd
			RGBForeColor(&skin_viscolor[0]);
			MoveTo(x,y);
			LineTo(x,y0);
		}else{ 				//case x is eaven
			for( ; y0>y; ){
				SetCPixel(x, y, &skin_viscolor[0]);	y++;
				SetCPixel(x, y, &skin_viscolor[1]);	y++;
			}
		}
		
		for( y=y0,c=2; y<58; y++, c++){
			SetCPixel(x, y, &skin_viscolor[c]);
		}
	}	
}

void skin_draw_spectrum(double value[])
{
	int i;
	int work[SPEC_WIDTH];
	
	for( i=0; i<SPEC_WIDTH; i++){ work[i]=0; }
	for( i=0; i<512; i++ ){
		work[i/8]+=value[i];
	}
	
	SetPortWindowPort(win.ref);
	/*back.red=  skin_viscolor[0][0]*256;
	back.green=skin_viscolor[0][1]*256;
	back.blue= skin_viscolor[0][2]*256;
	fore.red=  skin_viscolor[2][0]*256;
	fore.green=skin_viscolor[2][1]*256;
	fore.blue= skin_viscolor[2][2]*256;*/
	
	skin_draw_spectrum_fire(work);
	//skin_draw_spectrum_normal(work);
	//ctl->cmsg(CMSG_ERROR, VERB_NORMAL, "." );
}

void ctl_speana_data(double *val, int /*size*/) {

  /* 0 <= val[n] <= (AMP*NCOLOR) */
  /* AMP and NCOLOR are defined in soundspec.c */
  /* By default, AMP*NCOLOR = 1.0*64 */

  /* size = FFTSIZE/2 = 512 */
  /* FFTSIZE is also defined in soundspec.c */

  /* val[n] is the value of FFTed audio data */

#ifdef SUPPORT_SOUNDSPEC
  int i;
  int tx,x;
  double px;
  double s,a;
  int n;

  /*if ( isspeanaenabled )*/ {

    px=0.0;
    speana_buf[0] = (unsigned char)val[0];
    for ( i=1 ; i<SPE_W-1 ; i++ ) {
      s=0.0;
      n=0;
      tx=exp_hz_table[i];
      x=(int)px;

      do {
	a=val[x];
	s += a + (tx-x)*(val[x+1]-a);
	n++;
	x++;
      } while ( x<tx );

      s/=n;
      s*=16;
      if ( s<0 ) s=0;
      if ( s>=NCOLOR ) s=NCOLOR-1;
      speana_buf[i] = (unsigned char)(256*s/NCOLOR);
      px=tx;
    }
    speana_buf[SPE_W-1] = val[FFTSIZE/2-1];

    skin_draw_spectrum_fire(speana_buf);

  }
#endif /* SUPPORT_SOUNDSPEC */

  return;
}

static void initialize_exp_hz_table( void ) {
  int i;
  double r, x, w;
  
  w = (double)play_mode->rate * 0.5 / DEFAULT_ZOOM;
  r = exp(log(w) * (1.0/SPE_W));
  w = (FFTSIZE/2.0) / (w - 1.0);

  for(i = 0, x = 1.0; i <= SPE_W; i++, x *= r)
    exp_hz_table[i] = (x - 1.0) * w;

}

// *****************************************************
#pragma mark -

static int open_SkinWin()
		/*success-> return 0;*/
{
	OSErr	err;
	static RgnHandle	receiveRgn;

	open_window( &win, kSkinWinID);
	position_window(&win);
	err= skin_setup();
	if( err==noErr ) win.opened=true;

	if(gHasDragMgr){
		receiveRgn=NewRgn();
		if( receiveRgn ){
			GetWindowRegion(win.ref, kWindowContentRgn, receiveRgn);
			err=InstallTrackingHandler(NewDragTrackingHandlerProc(DragTrackingProc),
								(WindowPtr)win.ref, 0);
			if(err) ExitToShell();
			
			err=InstallReceiveHandler(NewDragReceiveHandlerProc(DragReceiveFunc),
							(WindowPtr)win.ref, 0);
			if(err) ExitToShell();
		}
	}
	skin_need_speana=false;
	initialize_exp_hz_table();
	return err;
}

/*static void Action(int btnid)
{
	switch( btnid ){
	case skin_btn_prev:
		mac_rc=RC_PREVIOUS;
		//mac_HandleControl();
		break;
	case skin_btn_play:
		mac_rc=RC_CONTINUE;
		//mac_HandleControl();
		break;
	case skin_btn_pause:
		mac_rc=RC_TOGGLE_PAUSE;
		//mac_HandleControl();
		break;
	case skin_btn_stop:
		mac_rc=RC_QUIT; ctl->current_time(0);
		//mac_HandleControl();
		break;
	case skin_btn_next:
		mac_rc=RC_NEXT;
		//mac_HandleControl();
		break;
	case skin_btn_eject:
		return;
	case skin_btn_shuffle:
		//gShuffle= !gShuffle;
		gShuffle=skin_buttons[skin_btn_shuffle].led;
		return;	 // don't call mac_HandleControl();
	case skin_btn_repeat:
		//skin_f_repeat= !skin_f_repeat;
		skin_f_repeat= skin_buttons[skin_btn_repeat].led;
		return; // don't call mac_HandleControl();
	case skin_btn_playlist:
		SHOW_WINDOW(mac_ListWindow);
		return;
	default: return;
	}
	mac_HandleControl();
}*/

/*static void HandleButton(int btnid, Point local)
{
	//int toggle_add=0;
	
	if( PtInRect(local, &skin_buttons[btnid].rect) ){
		//toggle_add=0;
		//if( btnid==skin_btn_shuffle && gShuffle ) toggle_add=2;
		//if( btnid==skin_btn_repeat && skin_f_repeat ) toggle_add=2;
		skin_update_button(&skin_buttons[btnid], 1);
		skin_redisp(skin_buttons[btnid].rect);
		Action(btnid);
		
		while(Button()){
			YieldToAnyThread();
		}
		
		//toggle_add=0;
		//if( btnid==skin_btn_shuffle && gShuffle ) toggle_add=2;
		//if( btnid==skin_btn_repeat && skin_f_repeat ) toggle_add=2;
		skin_update_button(&skin_buttons[btnid], 0);
		skin_redisp(skin_buttons[btnid].rect);
	}
}*/

static void click_SkinWin(Point local, short /*modifiers*/)
{
	static Point p;
		
	if( !win.opened ) return;
		
	p=local;
	NewThread(kCooperativeThread, xskin_jobs,
		         (void*)&p, 0, kCreateIfNeeded, 0, 0);
}

static void update_SkinWin()
{
	if( !win.opened ) return;
	if( bmpfile_changed_flag ){
		skin_remake_disp(panelrect);
		bmpfile_changed_flag=false;
	}
	skin_redisp(panelrect);
}

void goaway_SkinWin(MacWindow* macwin)
{
	goaway_default(macwin);
	skin_need_speana=false;
}

static int	message_SkinWin(int message, long param)
{
	if( !win.opened ) return 1;
	
	switch(message){
	case MW_SKIN_LOAD_BMP:
		skin_load_bmp((char*)param);
		break;
	case MW_SKIN_TOTAL_TIME:
		skin_put_time(0, 0);
		skin_posbar(0, param);
		break;
	case MW_SKIN_TIME:
		skin_put_time(param/60, param%60);
		skin_posbar(param, -1);
		break;
	case MW_SKIN_FILENAME:
		skin_filename((char*)param);
		break;
	}
	return -1;  //not supported
}

#endif /*TARGET_API_MAC_CARBON*/