/* 
    TiMidity++ -- MIDI to WAVE converter and player
    Copyright (C) 1999 Masanao Izumo <mo@goice.co.jp>
    Copyright (C) 1995 Tuukka Toivonen <tt@cgs.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	Macintosh interface for TiMidity
	by T.Nogami	<t-nogami@happy.email.ne.jp>
		
    mac_oms_a.c
    Macintosh QuickTime audio driver
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#ifdef ENABLE_OMS_AUDIO
#include <stdio.h>
#ifndef __WIN32__
#include <unistd.h>
#endif
#include <stdlib.h>

#ifndef NO_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#include "OMS.h"
#include "OMSDeviceMenu.h"

#include "timidity.h"
#include "common.h"
#include "instrum.h"
#include "playmidi.h"
#include "readmidi.h"
#include "output.h"
#include "controls.h"
#include "miditrace.h"
#include "wrd.h"

#include "mac_main.h"

static int open_output(void);
static void close_output(void);
static int output_data(char *buf, int32 count);
static int acntl(int request, void *arg);

static int wait_event_time(void *);
static void oms_play_event(void *);

static unsigned long	start_tic;
#if	0
static WindowPtr gWindow = nil;
#else
static DialogPtr gDialog = nil;
#endif
static OMSDeviceMenuH	gOutputMenu[2] = { nil, nil };
static short	gOutNodeItemNum[2] = { 1, 1 }, gOutNodeRefNum[2], gOutputPortRefNum;

#define dmp mac_oms_play_mode

PlayMode dmp =
{
    DEFAULT_RATE, PE_16BIT|PE_SIGNED, PF_MIDI_EVENT|PF_CAN_TRACE,
    -1,
    {0,0,0,0,0},
    "OMS output", 'o',
    "-",
    open_output,
    close_output,
    output_data,
    acntl
};

static void init_variable(void)
{
	start_tic = TickCount();
}

static pascal Boolean	dialogFilter(DialogPtr /*dialog*/, EventRecord *event, short * /*item*/)
{
	short winPart;
	WindowPtr whichWin;
	Point pt;

	switch (event->what) {
	case updateEvt:
		whichWin = (WindowPtr)event->message;
		SetPort(whichWin);
		BeginUpdate(whichWin);
		DrawOMSDeviceMenu(gOutputMenu[0]);
		DrawOMSDeviceMenu(gOutputMenu[1]);
		EndUpdate(whichWin);
		break;
	case mouseDown:
		winPart = FindWindow(event->where, &whichWin);
		switch (winPart) {
		case inContent:
			pt = event->where;
			SetPort(whichWin);
			GlobalToLocal(&pt);
			if (TestOMSDeviceMenu(gOutputMenu[0], pt)) {
				if (ClickOMSDeviceMenu(gOutputMenu[0])) {
					gOutNodeRefNum[0] = (*gOutputMenu[0])->selectedRefNum;
					gOutNodeItemNum[0] = (*gOutputMenu[0])->selectedItemNum;
				}
			}
			else if (TestOMSDeviceMenu(gOutputMenu[1], pt)) {
				if (ClickOMSDeviceMenu(gOutputMenu[1])) {
					gOutNodeItemNum[1] = (*gOutputMenu[1])->selectedItemNum;
				}
			}
			break;
		}
		break;
	case app4Evt:
		if (event->message & 0x01000000) {
			if (event->message & 0x00000001) {
				OMSResume('TIMI');
			} else {
				OMSSuspend('TIMI');
			}
		}
		break;
	}
	return 0;
}

static void	EventLoop(void)
{
	EventRecord theEvent;
	short winPart;
	WindowPtr whichWin;
	Point pt;
	int flag = 0;
	unsigned long delay;
	
	while (flag != 3) {
		if (WaitNextEvent(everyEvent, &theEvent, 1L, NULL)) {
			switch (theEvent.what) {
			case updateEvt:
				whichWin = (WindowPtr)theEvent.message;
				SetPort(whichWin);
				BeginUpdate(whichWin);
				DrawOMSDeviceMenu(gOutputMenu[0]);
				DrawOMSDeviceMenu(gOutputMenu[1]);
				EndUpdate(whichWin);
				break;
			case mouseDown:
				winPart = FindWindow(theEvent.where, &whichWin);
				switch (winPart) {
				case inGoAway:
					if (TrackGoAway(whichWin, theEvent.where))
						CloseWindow(whichWin);
					break;
				case inContent:
					pt = theEvent.where;
					SetPort(whichWin);
					GlobalToLocal(&pt);
					if (TestOMSDeviceMenu(gOutputMenu[0], pt)) {
						if (ClickOMSDeviceMenu(gOutputMenu[0])) {
							gOutNodeRefNum[0] = (*gOutputMenu[0])->selectedRefNum;
							gOutNodeItemNum[0] = (*gOutputMenu[0])->selectedItemNum;
							flag |= 1;
						}
					}
					else if (TestOMSDeviceMenu(gOutputMenu[1], pt)) {
						if (ClickOMSDeviceMenu(gOutputMenu[1])) {
							gOutNodeItemNum[1] = (*gOutputMenu[1])->selectedItemNum;
							flag |= 2;
						}
					}
					break;
				}
				break;
			case app4Evt:
				if (theEvent.message & 0x01000000) {
					if (theEvent.message & 0x00000001) {
						OMSResume('TIMI');
					} else {
						OMSSuspend('TIMI');
					}
				}
				break;
			}
		}
	}
	Delay(60, &delay);
	CloseWindow(whichWin);
}

static int open_output(void)
{
	short compatMode, itemHit;
	OMSErr err;
	Rect menuRect;

	if(dmp.fd == -1){
		// Find out what version of OMS is installed, if any.
		if(OMSVersion() == 0) 
			return -1;
		// Sign in to OMS
		err = OMSSignIn('TIMI', (long)LMGetCurrentA5(), LMGetCurApName(), nil, &compatMode);
		if (err)
			return -1;
		err = OMSAddPort('TIMI', 'out ', omsPortTypeOutput, NULL, 0L, &gOutputPortRefNum);
		if(err){
			OMSSignOut('TIMI');
			return -1;
		}
	}
#if	0
	gWindow = GetNewWindow(kOmsWinID, gWindow, (WindowPtr)-1L);
	SetPort(gWindow);
	SetRect(&menuRect, 30,22, 170,38);
	gOutputMenu[0] = NewOMSDeviceMenu(gOutputMenu[0], odmFrameBox, &menuRect, 
		omsIncludeOutputs + omsIncludeReal + omsIncludeVirtual, NULL);
	DrawOMSDeviceMenu(gOutputMenu[0]);
	MoveTo(10, 34);
	DrawString("\pA");
	SetRect(&menuRect, 30,52, 170,68);
	gOutputMenu[1] = NewOMSDeviceMenu(gOutputMenu[1], odmFrameBox, &menuRect, 
		omsIncludeOutputs + omsIncludeReal + omsIncludeVirtual, NULL);
	DrawOMSDeviceMenu(gOutputMenu[1]);
	MoveTo(11, 64);
	DrawString("\pB");
	ShowWindow(gWindow);
	EventLoop();
#else
	gDialog = GetNewDialog(300, gDialog, (DialogPtr)-1L);
	SetPort(gDialog);
	SetRect(&menuRect, 30,22, 170,38);
	gOutputMenu[0] = NewOMSDeviceMenu(gOutputMenu[0], odmFrameBox, &menuRect, 
		omsIncludeOutputs + omsIncludeReal + omsIncludeVirtual, NULL);
	DrawOMSDeviceMenu(gOutputMenu[0]);
	MoveTo(10, 34);
	DrawString("\pA");
	SetRect(&menuRect, 30,52, 170,68);
	gOutputMenu[1] = NewOMSDeviceMenu(gOutputMenu[1], odmFrameBox, &menuRect, 
		omsIncludeOutputs + omsIncludeReal + omsIncludeVirtual, NULL);
	DrawOMSDeviceMenu(gOutputMenu[1]);
	MoveTo(11, 64);
	DrawString("\pB");
	ShowWindow(gDialog);
	SetDialogDefaultItem(gDialog, ok);
	for(;itemHit != ok;){
		ModalDialog(NewModalFilterProc(dialogFilter), &itemHit);
	}
	CloseDialog(gDialog);
#endif
	dmp.fd = 1; //normaly opened flag
	return 0;
}

static int output_data(char *, int32)
{
 	// Never called
 	return 0;
}

static long current_tick(void)
{
	return TickCount() - start_tic;
}

static void close_output(void)
{
	if(dmp.fd == -1) return;
	OMSSignOut('TIMI');
	dmp.fd = -1;	//disabled
}

static void oms_ctl_note_event(int channel, int note, int velocity, Boolean on)
{
	CtlEvent ce;
	ce.type = CTLE_NOTE;
	ce.v1 = (on ? VOICE_ON : VOICE_FREE);
	ce.v2 = channel;
	ce.v3 = note;
	ce.v4 = velocity;
	if(ctl->trace_playing && !midi_trace.flush_flag)
		push_midi_trace_ce(ctl->event, &ce);
	else
		ctl->event(&ce);
}

static void ctl_timestamp(void)
{
	long secs;
	CtlEvent ce;
	static int last_secs = -1;

	secs = current_tick() / 60;
	if(secs == last_secs)
		return;
	ce.type = CTLE_CURRENT_TIME;
	ce.v1 = last_secs = secs;
	if(ctl->trace_playing && !midi_trace.flush_flag)
		push_midi_trace_ce(ctl->event, &ce);
	else
		ctl->event(&ce);
}

static void ctl_prog_event(int channel, int program)
{
	CtlEvent ce;
	ce.type = CTLE_PROGRAM;
	ce.v1 = channel;
	ce.v2 = program;
	ce.v3 = (long)"OMS";
	if(ctl->trace_playing && !midi_trace.flush_flag)
		push_midi_trace_ce(ctl->event, &ce);
	else
		ctl->event(&ce);
}

#define	OMS_WRITE_PACKET3(p0, p1, p2)	{ OMSMIDIPacket pkt;\
											pkt.flags = 0; pkt.len = 3; pkt.data[0] = p0 + ch;\
											pkt.data[1] = p1; pkt.data[2] = p2;\
											OMSWritePacket2(&pkt, gOutNodeRefNum[port], gOutputPortRefNum); }
#define	OMS_WRITE_PACKET2(p0, p1)		{ OMSMIDIPacket pkt;\
											pkt.flags = 0; pkt.len = 2; pkt.data[0] = p0 + ch;\
											pkt.data[1] = p1;\
											OMSWritePacket2(&pkt, gOutNodeRefNum[port], gOutputPortRefNum); }
#define	OMS_WRITE_CTL_PACKET(p0, p1)	{ OMSMIDIPacket pkt;\
											pkt.flags = 0; pkt.len = 3; pkt.data[0] = 0xb0 + ch;\
											pkt.data[1] = p0; pkt.data[2] = p1;\
											OMSWritePacket2(&pkt, gOutNodeRefNum[port], gOutputPortRefNum); }
#define	OMS_WRITE_GM_ON()	{ OMSMIDIPacket255 pkt; pkt.flags = 0; pkt.len = 6;\
								pkt.data[0] = 0xf0; pkt.data[1] = 0x7e; pkt.data[2] = 0x7f;\
								pkt.data[3] = 0x09; pkt.data[4] = 0x01; pkt.data[5] = 0xf7;\
								OMSWritePacket255(&pkt, gOutNodeRefNum[port], gOutputPortRefNum); }
#define	OMS_WRITE_GM_OFF()	{ OMSMIDIPacket255 pkt; pkt.flags = 0; pkt.len = 6;\
								pkt.data[0] = 0xf0; pkt.data[1] = 0x7e; pkt.data[2] = 0x7f;\
								pkt.data[3] = 0x09; pkt.data[4] = 0x02; pkt.data[5] = 0xf7;\
								OMSWritePacket255(&pkt, gOutNodeRefNum[port], gOutputPortRefNum); }
#define	OMS_WRITE_GS_ON()	{ OMSMIDIPacket255 pkt; pkt.flags = 0; pkt.len = 11;\
								pkt.data[0] = 0xf0; pkt.data[1] = 0x41; pkt.data[2] = 0x10;\
								pkt.data[3] = 0x42; pkt.data[4] = 0x12; pkt.data[5] = 0x40;\
								pkt.data[6] = 0x00; pkt.data[7] = 0x7f; pkt.data[8] = 0x00;\
								pkt.data[9] = 0x41; pkt.data[10] = 0xf7;\
								OMSWritePacket255(&pkt, gOutNodeRefNum[port], gOutputPortRefNum); }
#define	OMS_WRITE_GS_OFF()	{ OMSMIDIPacket255 pkt; pkt.flags = 0; pkt.len = 11;\
								pkt.data[0] = 0xf0; pkt.data[1] = 0x41; pkt.data[2] = 0x10;\
								pkt.data[3] = 0x42; pkt.data[4] = 0x12; pkt.data[5] = 0x40;\
								pkt.data[6] = 0x00; pkt.data[7] = 0x7f; pkt.data[8] = 0x7f;\
								pkt.data[9] = 0x42; pkt.data[10] = 0xf7;\
								OMSWritePacket255(&pkt, gOutNodeRefNum[port], gOutputPortRefNum); }
#define	OMS_WRITE_XG_ON()	{ OMSMIDIPacket255 pkt; pkt.flags = 0; pkt.len = 9;\
								pkt.data[0] = 0xf0; pkt.data[1] = 0x43; pkt.data[2] = 0x10;\
								pkt.data[3] = 0x4c; pkt.data[4] = 0x00; pkt.data[5] = 0x00;\
								pkt.data[6] = 0x7e; pkt.data[7] = 0x00; pkt.data[8] = 0xf7;\
								OMSWritePacket255(&pkt, gOutNodeRefNum[port], gOutputPortRefNum); }

static Boolean inited = 0;

static int wait_event_time(void *p)
{
	MidiEvent *ev = (MidiEvent *)p;
	int port, ch, rc;

	for(;;){
		if( ev->time - (current_tick()*play_mode->rate+30)/60 < 0 )
			return RC_NONE;
		trace_loop();
		YieldToAnyThread();
		rc = check_apply_control();
		if(RC_IS_SKIP_FILE(rc)){
			inited = 0;
			for(port = 0; port < 2; port++)
				for(ch = 0; ch < 16; ch++){
					OMS_WRITE_CTL_PACKET(123, ev->a);	// all note off
					OMS_WRITE_CTL_PACKET(120, ev->a);	// all sounds off
					OMS_WRITE_CTL_PACKET(121, ev->a);	// reset controllers
				}
			trace_flush();
			return rc;
		}
	}
}

static void oms_play_event(void *p)
{
	MidiEvent *ev = (MidiEvent *)p;
	int port, ch, i;

	port = (ev->channel>>4);
	ch = (ev->channel&15);
	switch(ev->type)
	{
	case ME_NOTEON:
		OMS_WRITE_PACKET3(0x90, ev->a, ev->b);
		oms_ctl_note_event(ch, ev->a, ev->b, true);
		break;
	case ME_NOTEOFF:
		OMS_WRITE_PACKET3(0x90, ev->a, 0);
		oms_ctl_note_event(ch, ev->a, ev->b, false);
		break;
	case ME_PROGRAM:
		OMS_WRITE_PACKET2(0xc0, ev->a);
		ctl_prog_event(ch, ev->a);
		channel[ch].program = ev->a;
		if(ISDRUMCHANNEL(ch))
			SET_CHANNELMASK(drumchannels, ch);
		break;
	/* MIDI Events */
	case ME_KEYPRESSURE:
		OMS_WRITE_PACKET3(0xa0, ev->a, ev->b);
		break;
	case ME_CHANNEL_PRESSURE:
		OMS_WRITE_PACKET2(0xd0, ev->a);
		break;
	case ME_PITCHWHEEL:
		OMS_WRITE_PACKET3(0xe0, ev->a, ev->b);
		break;
	/* Controls */
	case ME_TONE_BANK_LSB:
		OMS_WRITE_CTL_PACKET(32, ev->a);
		break;
	case ME_TONE_BANK_MSB:
		OMS_WRITE_CTL_PACKET(0, ev->a);
		break;
	case ME_MODULATION_WHEEL:
		OMS_WRITE_CTL_PACKET(1, ev->a);
		break;
	case ME_BREATH:
		OMS_WRITE_CTL_PACKET(2, ev->a);
		break;
	case ME_FOOT:
		OMS_WRITE_CTL_PACKET(4, ev->a);
		break;
	case ME_BALANCE:
		OMS_WRITE_CTL_PACKET(8, ev->a);
		break;
	case ME_MAINVOLUME:
		OMS_WRITE_CTL_PACKET(7, ev->a);
		break;
	case ME_PAN:
		OMS_WRITE_CTL_PACKET(10, ev->a);
		ctl_mode_event(CTLE_PANNING, 1, ch, ev->a);
		break;
	case ME_EXPRESSION:
		OMS_WRITE_CTL_PACKET(11, ev->a);
		ctl_mode_event(CTLE_EXPRESSION, 1, ch, ev->a);
		break;
	case ME_SUSTAIN:
		OMS_WRITE_CTL_PACKET(64, ev->a);
		ctl_mode_event(CTLE_SUSTAIN, 1, ch, ev->a);
		break;
	case ME_PORTAMENTO_TIME_MSB:
		OMS_WRITE_CTL_PACKET(5, ev->b);
		break;
	case ME_PORTAMENTO_TIME_LSB:
		OMS_WRITE_CTL_PACKET(37, ev->a);
		break;
	case ME_PORTAMENTO:
		OMS_WRITE_CTL_PACKET(65, ev->a);
		break;
	case ME_DATA_ENTRY_MSB:
		OMS_WRITE_CTL_PACKET(6, ev->a);
		break;
	case ME_REVERB_EFFECT:
		OMS_WRITE_CTL_PACKET(91, ev->a);
		break;
	case ME_TREMOLO_EFFECT:
		OMS_WRITE_CTL_PACKET(92, ev->a);
		break;
	case ME_CHORUS_EFFECT:
		OMS_WRITE_CTL_PACKET(93, ev->a);
		break;
	case ME_CELESTE_EFFECT:
		OMS_WRITE_CTL_PACKET(94, ev->a);
		break;
	case ME_PHASER_EFFECT:
		OMS_WRITE_CTL_PACKET(95, ev->a);
		break;
	case ME_RPN_INC:
		OMS_WRITE_CTL_PACKET(96, ev->a);
		break;
	case ME_RPN_DEC:
		OMS_WRITE_CTL_PACKET(97, ev->a);
		break;
	case ME_NRPN_LSB:
		OMS_WRITE_CTL_PACKET(98, ev->a);
		break;
	case ME_NRPN_MSB:
		OMS_WRITE_CTL_PACKET(99, ev->a);
		break;
	case ME_RPN_LSB:
		OMS_WRITE_CTL_PACKET(100, ev->a);
		break;
	case ME_RPN_MSB:
		OMS_WRITE_CTL_PACKET(101, ev->a);
		break;
	case ME_ALL_SOUNDS_OFF:
		OMS_WRITE_CTL_PACKET(120, ev->a);
		break;
	case ME_RESET_CONTROLLERS:
		OMS_WRITE_CTL_PACKET(121, ev->a);
		ctl_mode_event(CTLE_VOLUME, 1, ch, 0);
		ctl_mode_event(CTLE_EXPRESSION, 1, ch, 127);
		ctl_mode_event(CTLE_SUSTAIN, 1, ch, 0);
		ctl_mode_event(CTLE_MOD_WHEEL, 1, ch, 0);
		ctl_mode_event(CTLE_PITCH_BEND, 1, ch, 0x2000);
		ctl_prog_event(ch, ev->a);
		ctl_mode_event(CTLE_CHORUS_EFFECT, 1, ch, 0);
		ctl_mode_event(CTLE_REVERB_EFFECT, 1, ch, 0);
		break;
	case ME_ALL_NOTES_OFF:
		OMS_WRITE_CTL_PACKET(123, ev->a);
		break;
	case ME_MONO:
		OMS_WRITE_CTL_PACKET(126, ev->a);
		break;
	case ME_POLY:
		OMS_WRITE_CTL_PACKET(127, ev->a);
		break;
	case ME_SOSTENUTO:
		OMS_WRITE_CTL_PACKET(66, ev->a);
		break;
	case ME_SOFT_PEDAL:
		OMS_WRITE_CTL_PACKET(67, ev->a);
		break;
	case ME_HARMONIC_CONTENT:
		OMS_WRITE_CTL_PACKET(71, ev->a);
		break;
	case ME_RELEASE_TIME:
		OMS_WRITE_CTL_PACKET(72, ev->a);
		break;
	case ME_ATTACK_TIME:
		OMS_WRITE_CTL_PACKET(73, ev->a);
		break;
	case ME_BRIGHTNESS:
		OMS_WRITE_CTL_PACKET(74, ev->a);
		break;
	case ME_PORTAMENTO_CONTROL:
		OMS_WRITE_CTL_PACKET(84, ev->a);
		break;
	/* TiMidity Extensionals */
	case ME_RANDOM_PAN:
		break;
	case ME_SET_PATCH:
		break;
	case ME_TEMPO:
		break;
	case ME_CHORUS_TEXT:
	case ME_LYRIC:
	case ME_MARKER:
	case ME_INSERT_TEXT:
	case ME_TEXT:
	case ME_KARAOKE_LYRIC:
		i = ev->a | ((int)ev->b << 8);
		ctl_mode_event(CTLE_LYRIC, 1, i, 0);
		break;
	case ME_MASTER_VOLUME:
		break;
	case ME_PATCH_OFFS:
		break;
	case ME_RESET:
		if(ev->a == GM_SYSTEM_MODE){
			OMS_WRITE_GM_ON();
		}
		else if(ev->a == GS_SYSTEM_MODE){
			OMS_WRITE_GS_ON();
		}
		else if(ev->a == XG_SYSTEM_MODE){
			OMS_WRITE_XG_ON();
		}
		break;
	case ME_WRD:
		push_midi_trace2(wrd_midi_event, ch, ev->a | (ev->b << 8));
		break;
	case ME_DRUMPART:
		break;
	case ME_KEYSHIFT:
		break;
	case ME_NOTE_STEP:
		break;
	case ME_EOT:
		inited = 0;
		trace_flush();
		break;
	}
	if(ev->type != ME_EOT)
		ctl_timestamp();
}

static int acntl(int request, void *arg)
{
	int rc;

    switch(request){
    case PM_REQ_MIDI:
    	if(!inited){
			inited = 1;
			init_variable();
		}
		rc = wait_event_time(arg);
		if(RC_IS_SKIP_FILE(rc))
			return rc;
    	oms_play_event(arg);
    	return RC_NONE;
	case PM_REQ_DISCARD:
	case PM_REQ_FLUSH:
	case PM_REQ_PLAY_START:
		init_variable();
		return 0;
    }
    return -1;
}
#endif //ENABLE_OMS_AUDIO
